import pathlib

DEFAULT_RESULT_DIR = pathlib.Path("./.minto_experiments")
DEFAULT_PROBLEM_NAMES = [
    "bin-packing",
    "knapsack",
    "nurse-scheduling",
    "travelling-salesman",
    "travelling-salesman-with-time-windows",
]
